;; Generated package description from flx-ido.el  -*- no-byte-compile: t -*-
(define-package "flx-ido" "0.6.2" "flx integration for ido" '((flx "0.1") (cl-lib "0.3")) :commit "29e3664b75618847eaaf821713a74eb3c897a124" :url "https://github.com/lewang/flx")
